﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;

using JigLibSDX;
using JigLibSDX.Collision;
using JigLibSDX.Geometry;
using JigLibSDX.Math;
using JigLibSDX.Physics;
using JigLibSDX.Utils;

using SDX = SlimDX;
using SD9 = SlimDX.Direct3D9;

using DemoFramework;

namespace Demo_2
{
    public partial class Demo : Form
    {
        #region Non-Physics Code: Attributes
        //directx
        private SD9.Device _graphigsDevice;
        private SD9.Direct3D _direct3D;

        //screen size
        private const int _WIDTH = 640;
        private const int _HEIGHT = 480;

        //cam
        private Camera _camera;

        //input
        private BasicInput _input;

        //timing
        private Stopwatch _stopwatch;
        private float _secondsPerFrame = 0.016666f;
        #endregion

        private PhysicsSystem _physics;

        //Sphere
        private SD9.Mesh _sphereMesh;
        private CollisionSkin _sphereSkin;
        private Body _sphereBody;
        private SDX.Vector3 _sphereCenterOfMass;

        //World
        private SD9.Mesh _worldMesh;
        private CollisionSkin _worldSkin;
        private Body _worldBody;
        private SDX.Vector3 _worldCenterOfMass;

        #region Non-Physics Code: Basic Init, Update & Terminate Handling
        #region Init
        public Demo()
        {
            InitializeComponent();
            SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
            this.Width = _WIDTH;
            this.Height = _HEIGHT;

            _camera = new Camera(new SDX.Vector3(0f, 2f, -10f), SDX.Vector3.Zero, SDX.Vector3.UnitY, _WIDTH, _HEIGHT, (float)Math.PI / 2f, 0.001f, 100f);

            InitializeGraphics();
            InitializeLight();
            InitializePhysics();
            InitializePhysicsObjects();

            LoadContent();

            _input = new BasicInput(this.Handle);
            _stopwatch = new Stopwatch();
        }
        #endregion

        #region Form Closing
        private void Demo_FormClosing(object sender, FormClosingEventArgs e)
        {
            UnloadContent();
        }
        #endregion

        #region Update & Draw Controller
        private void Demo_Paint(object sender, PaintEventArgs e)
        {
            if (_stopwatch.IsRunning)
            {
                float elapsedSeconds = _stopwatch.ElapsedMilliseconds * 0.001f;

                if (elapsedSeconds >= _secondsPerFrame)
                {
                    _stopwatch.Reset();
                    _stopwatch.Start();

                    Update(elapsedSeconds);
                    Draw(elapsedSeconds);
                }
            }
            else
            {
                _stopwatch.Start();
            }

            this.Invalidate();
        }
        #endregion

        #region Init Graphics
        private void InitializeGraphics()
        {
            SD9.PresentParameters presentParams = new SD9.PresentParameters();
            presentParams.BackBufferWidth = _WIDTH;
            presentParams.BackBufferHeight = _HEIGHT;
            presentParams.BackBufferFormat = SD9.Format.A8R8G8B8;
            presentParams.DeviceWindowHandle = this.Handle;

            _direct3D = new SD9.Direct3D();
            _graphigsDevice = new SD9.Device(_direct3D, 0, SD9.DeviceType.Hardware, this.Handle, SD9.CreateFlags.HardwareVertexProcessing, presentParams);

            _graphigsDevice.SetRenderState(SD9.RenderState.AlphaBlendEnable, true);
            _graphigsDevice.SetRenderState(SD9.RenderState.SourceBlend, SD9.Blend.SourceAlpha);
            _graphigsDevice.SetRenderState(SD9.RenderState.DestinationBlend, SD9.Blend.BothInverseSourceAlpha);
            _graphigsDevice.SetRenderState(SD9.RenderState.Lighting, true);
            _graphigsDevice.SetRenderState(SD9.RenderState.CullMode, SD9.Cull.None);
            _graphigsDevice.SetRenderState(SD9.RenderState.FillMode, SD9.FillMode.Solid);
            _graphigsDevice.SetRenderState(SD9.RenderState.NormalizeNormals, true);
        }
        #endregion

        #region Init Lights
        private void InitializeLight()
        {
            //light 1
            SDX.Vector3 direction = new SDX.Vector3(-1f, 0f, -1f);
            direction.Normalize();

            SD9.Light light = new SD9.Light();
            light.Type = SD9.LightType.Directional;
            light.Direction = direction;
            light.Diffuse = Color.White;
            light.Ambient = Color.White;

            _graphigsDevice.SetLight(0, light);
            _graphigsDevice.EnableLight(0, true);

            //light 2
            direction = new SDX.Vector3(1f, 0f, 1f);
            direction.Normalize();

            light = new SD9.Light();
            light.Type = SD9.LightType.Directional;
            light.Direction = direction;
            light.Diffuse = Color.White;
            light.Ambient = Color.White;

            _graphigsDevice.SetLight(1, light);
            _graphigsDevice.EnableLight(1, true);
        }
        #endregion
        #endregion

        private void InitializePhysics()
        {
            _physics = new PhysicsSystem();

            _physics.CollisionSystem = new CollisionSystemSAP();
            _physics.CollisionSystem.UseSweepTests = true;

            /// I'm using PhysicsSystem.Solver.Accumulated because it's the best looking solver.
            /// This might be slow if you tend to have worlds with many complex objects.
            _physics.SolverType = PhysicsSystem.Solver.Accumulated;

            /// Improves speed and prevents jiggling.
            _physics.EnableFreezing = true;

            _physics.NumCollisionIterations = 10;
            _physics.NumContactIterations = 10;
            _physics.NumPenetrationRelaxtionTimesteps = 15;

            //Set Gravity
            _physics.Gravity = new SDX.Vector3(0f, -10f, 0f);
        }

        private void InitializePhysicsObjects()
        {
            CreateSphere(out _sphereBody, out _sphereSkin, out _sphereCenterOfMass);
            _sphereBody.Position = new SDX.Vector3(-7f, 5f, 0f);

            CreateWorld(out _worldBody, out _worldSkin, out _worldCenterOfMass);
            _worldBody.Position = new SDX.Vector3(0f, 0f, 0f);
            _worldBody.Immovable = true;
        }

        #region Non-Physics Code: Load & Unload Content
        private void LoadContent()
        {
            _sphereMesh = SD9.Mesh.FromFile(_graphigsDevice, @"Content\Meshes\sphere.x", SD9.MeshFlags.Managed);
            _sphereMesh.OptimizeInPlace(SD9.MeshOptimizeFlags.VertexCache | SD9.MeshOptimizeFlags.Compact | SD9.MeshOptimizeFlags.AttributeSort);

            _worldMesh = SD9.Mesh.FromFile(_graphigsDevice, @"Content\Meshes\world5.x", SD9.MeshFlags.Managed);
            _worldMesh.OptimizeInPlace(SD9.MeshOptimizeFlags.VertexCache | SD9.MeshOptimizeFlags.Compact | SD9.MeshOptimizeFlags.AttributeSort);
        }

        private void UnloadContent()
        {
            _sphereMesh = null;
        }
        #endregion

        private void Update(float elapsedSeconds)
        {
            #region Non-Physics Code: Input Update
            _input.Update();
            #endregion

            //Apply forces
            // ...

            //Update Physics
            // Use stable steps (small steps are good - big steps are bad).
            // noone (JigLibX author) suggests you use a maximum step of 1/60.
            // 1/50 worked well for me, so I guess it's ok to use 1/50 :).
            if (elapsedSeconds < 0.02f)
            {
                _physics.Update(elapsedSeconds);
            }
            else
            {
                /// Ok, your game is running slowly. There are still some limits in computing.
                /// Use less complexity, objects or tell the user to lower the rendering settings.
                _physics.Update(0.02f);
            }


            #region Non-Physics: Note
            //Update Positions
            // ... for example: Camera or graphics positions that depend on objects from the physics world.
            _camera.Position = new SlimDX.Vector3(_sphereBody.Position.X - 2f, _sphereBody.Position.Y +1f, _sphereBody.Position.Z - 5f);
            _camera.Target = _sphereBody.Position;
            #endregion
        }

        #region Non-Physics Code: Draw
        private void Draw(float elapsedSeconds)
        {
            _graphigsDevice.Clear(SD9.ClearFlags.Target | SD9.ClearFlags.ZBuffer, Color.CornflowerBlue, 1.0f, 0);
            _graphigsDevice.BeginScene();

            _graphigsDevice.SetTransform(SD9.TransformState.View, _camera.View);
            _graphigsDevice.SetTransform(SD9.TransformState.Projection, _camera.Projection);

            _graphigsDevice.SetTransform(SD9.TransformState.World, _sphereBody.Orientation *
                                                                   SDX.Matrix.Translation(_sphereBody.Position - _sphereCenterOfMass));

            SD9.ExtendedMaterial[] extMat = _sphereMesh.GetMaterials();
            _graphigsDevice.Material = extMat[0].MaterialD3D;
            _sphereMesh.DrawSubset(0);

            _graphigsDevice.SetTransform(SD9.TransformState.World, _worldBody.Orientation *
                                                                   SDX.Matrix.Translation(_worldBody.Position - _worldCenterOfMass));

            extMat = _worldMesh.GetMaterials();
            _graphigsDevice.Material = extMat[0].MaterialD3D;
            _worldMesh.DrawSubset(0);

            _graphigsDevice.EndScene();
            _graphigsDevice.Present();
        }
        #endregion


        // The following functions were generated with the CollisionSkin Editor, which is part of JigLibSDX.
        private void CreateSphere(out JigLibSDX.Physics.Body body, out JigLibSDX.Collision.CollisionSkin skin, out SDX.Vector3 centerOfMass)
        {
            #region Header
            // Variables
            JigLibSDX.Collision.MaterialProperties materialProperties;
            JigLibSDX.Geometry.PrimitiveProperties primitiveProperties;
            JigLibSDX.Geometry.Sphere sphere;

            #region Mass Variables
            float mass;
            SDX.Matrix inertiaTensor;
            SDX.Matrix inertiaTensorCoM;
            #endregion

            // Create Skin & Body
            body = new JigLibSDX.Physics.Body();
            skin = new JigLibSDX.Collision.CollisionSkin(body);
            body.CollisionSkin = skin;
            #endregion

            #region Primitive Properties
            primitiveProperties = new JigLibSDX.Geometry.PrimitiveProperties();
            primitiveProperties.MassDistribution = JigLibSDX.Geometry.PrimitiveProperties.MassDistributionEnum.Shell;
            primitiveProperties.MassType = JigLibSDX.Geometry.PrimitiveProperties.MassTypeEnum.Mass;
            primitiveProperties.MassOrDensity = 2f;
            #endregion

            #region Primitive 0
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.8f;
            materialProperties.DynamicRoughness = 0.5f;
            materialProperties.Elasticity = 3f;

            // Primitive:
            sphere = new JigLibSDX.Geometry.Sphere(new SDX.Vector3(0f, 0f, 0f), 1f);
            skin.AddPrimitive(sphere, materialProperties);
            #endregion

            #region Footer
            // Extract Mass Properties
            skin.GetMassProperties(primitiveProperties, out mass, out centerOfMass, out inertiaTensor, out inertiaTensorCoM);

            // Set Mass Properties
            body.BodyInertia = inertiaTensorCoM;
            body.Mass = mass;

            // Sync Body & Skin
            body.MoveTo(SDX.Vector3.Zero, SDX.Matrix.Identity);
            skin.ApplyLocalTransform(new JigLibSDX.Math.Transform(-centerOfMass, SDX.Matrix.Identity));

            // Enable Body
            body.EnableBody();
            #endregion
        }

        private void CreateWorld(out JigLibSDX.Physics.Body body, out JigLibSDX.Collision.CollisionSkin skin, out SDX.Vector3 centerOfMass)
        {
            #region Header
            // Variables
            JigLibSDX.Collision.MaterialProperties materialProperties;
            JigLibSDX.Geometry.PrimitiveProperties primitiveProperties;
            JigLibSDX.Geometry.Box box;
            JigLibSDX.Geometry.Capsule capsule;

            #region Mass Variables
            float mass;
            SDX.Matrix inertiaTensor;
            SDX.Matrix inertiaTensorCoM;
            #endregion

            // Create Skin & Body
            body = new JigLibSDX.Physics.Body();
            skin = new JigLibSDX.Collision.CollisionSkin(body);
            body.CollisionSkin = skin;
            #endregion

            #region Primitive Properties
            primitiveProperties = new JigLibSDX.Geometry.PrimitiveProperties();
            primitiveProperties.MassDistribution = JigLibSDX.Geometry.PrimitiveProperties.MassDistributionEnum.Solid;
            primitiveProperties.MassType = JigLibSDX.Geometry.PrimitiveProperties.MassTypeEnum.Mass;
            primitiveProperties.MassOrDensity = 1000f;
            #endregion

            #region Primitive 0
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(-4.004753f, 1.458591f, -0.3994086f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(-1.083869E-09f, -1.083869E-09f, -0.02479604f, 0.9996926f)),
                new SDX.Vector3(1.995065f, 0.199328f, 0.7988173f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 1
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(-2.999662f, 0.9009169f, -0.4000822f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(1.999599f, 0.1980295f, 0.8001644f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 2
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(-1.99366f, 0.4026157f, -0.3974871f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(1.995773f, 0.1967787f, 0.7949742f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 3
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(-0.001073658f, -0.59433f, -0.3974871f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(1.997852f, 0.1949649f, 0.7949742f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 4
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(1.000916f, -1.089729f, -0.3993921f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(1.999281f, 0.1974811f, 0.8018563f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 5
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(-0.9974293f, -0.09744276f, -0.397487f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(1.994859f, 0.1948855f, 0.794974f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 6
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(3.676989f, -2.004423f, -1.629475f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(3.647572f, 0.1946911f, 3.25895f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 7
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            capsule = new JigLibSDX.Geometry.Capsule(new SDX.Vector3(7.641746f, -0.7274688f, -1.424341f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                0.4713295f,
                2.871125f);

            skin.AddPrimitive(capsule, materialProperties);
            #endregion

            #region Primitive 8
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(3.681396f, -1.808074f, -1.424021f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(0.2227128f, 0.4051367f, 2.848043f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 9
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(7.099108f, -1.81018f, -1.424022f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(0.2253551f, 0.4077798f, 2.848043f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 10
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(3.681893f, -1.817404f, 1.428578f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(3.628803f, 0.4174177f, 0.2030057f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Primitive 11
            // MaterialProperties:
            materialProperties = new JigLibSDX.Collision.MaterialProperties();
            materialProperties.StaticRoughness = 0.5f;
            materialProperties.DynamicRoughness = 0.35f;
            materialProperties.Elasticity = 0.25f;

            // Primitive:
            box = new JigLibSDX.Geometry.Box(new SDX.Vector3(3.680999f, -1.81398f, -1.65038f),
                SDX.Matrix.RotationQuaternion(new SDX.Quaternion(0f, 0f, 0f, 1f)),
                new SDX.Vector3(3.639106f, 0.4127079f, 0.2405989f));

            skin.AddPrimitive(box, materialProperties);
            #endregion

            #region Footer
            // Extract Mass Properties
            skin.GetMassProperties(primitiveProperties, out mass, out centerOfMass, out inertiaTensor, out inertiaTensorCoM);

            // Set Mass Properties
            body.BodyInertia = inertiaTensorCoM;
            body.Mass = mass;

            // Sync Body & Skin
            body.MoveTo(SDX.Vector3.Zero, SDX.Matrix.Identity);
            skin.ApplyLocalTransform(new JigLibSDX.Math.Transform(-centerOfMass, SDX.Matrix.Identity));

            // Enable Body
            body.EnableBody();
            #endregion
        }
    }
}
